import logging
from typing import Dict, List, Optional

import boto3
from mypy_boto3_ec2.client import EC2Client

from awspub import exceptions
from awspub.context import Context

logger = logging.getLogger(__name__)


class Snapshot:
    """
    Handle EC2 Snapshot API interaction
    """

    def __init__(self, context: Context):
        self._ctx: Context = context

    def _get(self, ec2client: EC2Client, snapshot_name: str) -> Optional[str]:
        """
        Get the snapshot id for the given name or None

        :param ec2client: EC2 client for a specific region
        :type ec2client: EC2Client
        :param snapshot_name: the Snapshot name
        :type snapshot_name: str
        :return: Either None or a snapshot-id
        :rtype: Optional[str]
        """
        resp = ec2client.describe_snapshots(
            Filters=[
                {
                    "Name": "tag:Name",
                    "Values": [
                        snapshot_name,
                    ],
                },
                {
                    "Name": "status",
                    "Values": [
                        "pending",
                        "completed",
                    ],
                },
            ],
            OwnerIds=["self"],
        )
        if len(resp.get("Snapshots", [])) == 1:
            return resp["Snapshots"][0]["SnapshotId"]
        elif len(resp.get("Snapshots", [])) == 0:
            return None
        else:
            raise exceptions.MultipleSnapshotsException(
                f"Found {len(resp.get('Snapshots', []))} snapshots with "
                f"name '{snapshot_name}' in region {ec2client.meta.region_name}"
            )

    def _get_import_snapshot_task(self, ec2client: EC2Client, snapshot_name: str) -> Optional[str]:
        """
        Get a import snapshot task for the given name

        :param ec2client: EC2 client for a specific region
        :type ec2client: EC2Client
        :param snapshot_name: the Snapshot name
        :type snapshot_name: str
        :return: Either None or a import-snapshot-task-id
        :rtype: Optional[str]
        """
        resp = ec2client.describe_import_snapshot_tasks(
            Filters=[
                {
                    "Name": "tag:Name",
                    "Values": [
                        snapshot_name,
                    ],
                }
            ]
        )
        # API doesn't support filters by status so filter here
        tasks: List = resp.get("ImportSnapshotTasks", [])
        # we already know here that the snapshot does not exist (checked in create() before calling this
        # function). so ignore "deleted" or "completed" tasks here
        # it might happen (for whatever reason) that a task got completed but the snapshot got deleted
        # afterwards. In that case a "completed" task for the given snapshot_name exists but
        # that doesn't help so ignore it
        tasks = [t for t in tasks if t["SnapshotTaskDetail"]["Status"] not in ["deleted", "completed"]]
        if len(tasks) == 1:
            return tasks[0]["ImportTaskId"]
        elif len(tasks) == 0:
            return None
        else:
            raise exceptions.MultipleImportSnapshotTasksException(
                f"Found {len(tasks)} import snapshot tasks with "
                f"name '{snapshot_name}' in region {ec2client.meta.region_name}"
            )

    def create(self, ec2client: EC2Client, snapshot_name: str) -> str:
        """
        Create a EC2 snapshot with the given name
        If the snapshot already exists, just return the snapshot-id for the existing snapshot.

        :param ec2client: EC2 client for a specific region
        :type ec2client: EC2Client
        :param snapshot_name: the Snapshot name
        :type snapshot_name: str
        :return: a snapshot-id
        :rtype: str
        """
        # does a snapshot with the given name already exists?
        snap_id: Optional[str] = self._get(ec2client, snapshot_name)
        if snap_id:
            logger.info(f"snapshot with name '{snapshot_name}' already exists in region {ec2client.meta.region_name}")
            return snap_id

        logger.info(
            f"Create snapshot from bucket '{self._ctx.conf['s3']['bucket_name']}' "
            f"for '{snapshot_name}' in region  {ec2client.meta.region_name}"
        )

        # extend tags
        tags = self._ctx.tags
        tags.append({"Key": "Name", "Value": snapshot_name})

        # does a import snapshot task with the given name already exist?
        import_snapshot_task_id: Optional[str] = self._get_import_snapshot_task(ec2client, snapshot_name)
        if import_snapshot_task_id:
            logger.info(
                f"import snapshot task ({import_snapshot_task_id}) with "
                f"name '{snapshot_name}' exists in region {ec2client.meta.region_name}"
            )
        else:
            resp = ec2client.import_snapshot(
                Description="Import ",
                DiskContainer={
                    "Description": "",
                    "Format": "vmdk",
                    "UserBucket": {
                        "S3Bucket": self._ctx.conf["s3"]["bucket_name"],
                        "S3Key": self._ctx.source_sha256,
                    },
                },
                TagSpecifications=[
                    {"ResourceType": "import-snapshot-task", "Tags": tags},
                ],
            )
            import_snapshot_task_id = resp["ImportTaskId"]

        logger.info(
            f"Waiting for snapshot import task (id: {import_snapshot_task_id}) "
            f"in region {ec2client.meta.region_name} ..."
        )

        waiter_import = ec2client.get_waiter("snapshot_imported")
        waiter_import.wait(ImportTaskIds=[import_snapshot_task_id], WaiterConfig={"Delay": 30, "MaxAttempts": 90})

        task_details = ec2client.describe_import_snapshot_tasks(ImportTaskIds=[import_snapshot_task_id])
        snapshot_id = task_details["ImportSnapshotTasks"][0]["SnapshotTaskDetail"]["SnapshotId"]

        # create tags before waiting for completion so the tags are already there
        ec2client.create_tags(Resources=[snapshot_id], Tags=tags)

        waiter_completed = ec2client.get_waiter("snapshot_completed")
        waiter_completed.wait(SnapshotIds=[snapshot_id], WaiterConfig={"Delay": 30, "MaxAttempts": 60})

        logger.info(f"Snapshot import as '{snapshot_id}' in region {ec2client.meta.region_name} done")
        return snapshot_id

    def _copy(self, snapshot_name: str, source_region: str, destination_region: str) -> str:
        """
        Copy a EC2 snapshot for the given context to the destination region
        NOTE: we don't wait for the snapshot to complete here!

        :param snapshot_name: the Snapshot name to copy
        :type snapshot_name: str
        :param source_region: a region to copy the snapshot from
        :type source_region: str
        :param destination_region: a region to copy the snapshot to
        :type destionation_region: str

        :return: the existing or created snapshot-id
        :rtype: str
        """

        # does the snapshot with that name already exist in the destination region?
        ec2client_dest: EC2Client = boto3.client("ec2", region_name=destination_region)
        snapshot_id: Optional[str] = self._get(ec2client_dest, snapshot_name)
        if snapshot_id:
            logger.info(
                f"snapshot with name '{snapshot_name}' already "
                f"exists ({snapshot_id}) in destination region {ec2client_dest.meta.region_name}"
            )
            return snapshot_id

        ec2client_source: EC2Client = boto3.client("ec2", region_name=source_region)
        source_snapshot_id: Optional[str] = self._get(ec2client_source, snapshot_name)
        if not source_snapshot_id:
            raise ValueError(
                f"Can not find source snapshot with name '{snapshot_name}' "
                f"in region {ec2client_source.meta.region_name}"
            )

        logger.info(f"Copy snapshot {source_snapshot_id} from " f"{source_region} to {destination_region}")
        # extend tags
        tags = self._ctx.tags
        tags.append({"Key": "Name", "Value": snapshot_name})
        resp = ec2client_dest.copy_snapshot(
            SourceRegion=source_region,
            SourceSnapshotId=source_snapshot_id,
            TagSpecifications=[{"ResourceType": "snapshot", "Tags": tags}],
        )

        # note: we don't wait for the snapshot to complete here!
        return resp["SnapshotId"]

    def copy(self, snapshot_name: str, source_region: str, destination_regions: List[str]) -> Dict[str, str]:
        """
        Copy a snapshot to multiple regions

        :param snapshot_name: the Snapshot name to copy
        :type snapshot_name: str
        :param source_region: a region to copy the snapshot from
        :type source_region: str
        :param destination_regions: a list of regions to copy the snaphot to
        :type destionation_regions: List[str]
        :return: a dict with region/snapshot-id mapping for the newly copied snapshots
        :rtype: Dict[str, str] where the key is a region name and the value a snapshot-id
        """
        snapshot_ids: Dict[str, str] = dict()
        for destination_region in destination_regions:
            snapshot_ids[destination_region] = self._copy(snapshot_name, source_region, destination_region)

        logger.info(f"Waiting for {len(snapshot_ids)} snapshots to appear in the destination regions ...")
        for destination_region, snapshot_id in snapshot_ids.items():
            ec2client_dest = boto3.client("ec2", region_name=destination_region)
            waiter = ec2client_dest.get_waiter("snapshot_completed")
            logger.info(f"Waiting for {snapshot_id} in {ec2client_dest.meta.region_name} to complete ...")
            waiter.wait(SnapshotIds=[snapshot_id], WaiterConfig={"Delay": 30, "MaxAttempts": 90})

        return snapshot_ids



import hashlib
import logging
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

import boto3
from mypy_boto3_ec2.client import EC2Client
from mypy_boto3_ssm import SSMClient

from awspub import exceptions
from awspub.context import Context
from awspub.image_marketplace import ImageMarketplace
from awspub.s3 import S3
from awspub.snapshot import Snapshot

logger = logging.getLogger(__name__)


@dataclass
class _ImageInfo:
    """
    Information about a image from EC2
    """

    image_id: str
    snapshot_id: Optional[str]


class ImageVerificationErrors(str, Enum):
    """
    Possible errors for image verification
    """

    NOT_EXIST = "image does not exist"
    STATE_NOT_AVAILABLE = "image not available"
    ROOT_DEVICE_TYPE = "root device type mismatch"
    ROOT_DEVICE_VOLUME_TYPE = "root device volume type mismatch"
    ROOT_DEVICE_VOLUME_SIZE = "root device volume size mismatch"
    ROOT_DEVICE_SNAPSHOT_NOT_COMPLETE = "root device snapshot not complete"
    BOOT_MODE = "boot mode mismatch"
    TAGS = "tags mismatch"
    TPM_SUPPORT = "tpm support mismatch"
    IMDS_SUPPORT = "imds support mismatch"
    BILLING_PRODUCTS = "billing products mismatch"


class Image:
    """
    Handle EC2 Image/AMI API interaction
    """

    def __init__(self, context: Context, image_name: str):
        self._ctx: Context = context
        self._image_name: str = image_name
        self._image_regions: List[str] = []

        if self._image_name not in self._ctx.conf["images"].keys():
            raise ValueError(f"image '{self._image_name}' not found in context configuration")

        self._snapshot: Snapshot = Snapshot(context)
        self._s3: S3 = S3(context)

    def __repr__(self):
        return f"<{self.__class__} :'{self.image_name}' (snapshot name: {self.snapshot_name})"

    @property
    def conf(self) -> Dict[str, Any]:
        """
        The configuration for the current image (based on "image_name") from context
        """
        return self._ctx.conf["images"][self._image_name]

    @property
    def image_name(self) -> str:
        """
        Get the image name
        """
        return self._image_name

    @property
    def snapshot_name(self) -> str:
        """
        Get the snapshot name which is a sha256 hexdigest

        The snapshot name is the sha256 hexdigest of the source file given in the source->path
        configuration option.

        if the "separate_snapshot" config option is set to True, the snapshot name is
        sha256 hexdigest of the source file given in the source->path conf option and then
        the sha256 hexdigest of the image-name appended and then the sha256 hexdigest
        calculated of this concatenated string.

        if the "billing_products" config option is set, the snapshot name is
        sha256 hexdigest of the source file given in the source->path conf option and then
        the sha256 hexdigest of each entry in the billing_products appended and then the sha256 hexdigest
        calculated of this concatenated string.

        Note that both options ("separate_snapshot" and "billing_products") can be combined
        and the snapshot calculation steps would be combined, too.
        """
        s_name = self._ctx.source_sha256
        if self.conf["separate_snapshot"] is True:
            s_name += hashlib.sha256(self.image_name.encode("utf-8")).hexdigest()

        if self.conf["billing_products"]:
            for bp in self.conf["billing_products"]:
                s_name += hashlib.sha256(bp.encode("utf-8")).hexdigest()

        # in the separate_snapshot and billing_products had no effect, don't do another sha256 of
        # the source_sha256 to simplify things
        if s_name == self._ctx.source_sha256:
            return s_name

        # do a sha256 of the concatenated string
        return hashlib.sha256(s_name.encode("utf-8")).hexdigest()

    @property
    def image_regions(self) -> List[str]:
        """
        Get the image regions. Either configured in the image configuration
        or all available regions
        """
        if not self._image_regions:
            if self.conf["regions"]:
                self._image_regions = self.conf["regions"]
            else:
                ec2client: EC2Client = boto3.client("ec2", region_name=self._s3.bucket_region)
                resp = ec2client.describe_regions()
                self._image_regions = [r["RegionName"] for r in resp["Regions"]]
        return self._image_regions

    @property
    def _tags(self):
        """
        Get the tags for this image (common tags + image specific tags)
        image specific tags override common tags
        """
        tags = []
        # the common tags
        tags_dict = self._ctx.tags_dict
        # the image specific tags
        tags_dict.update(self.conf.get("tags", {}))
        for name, value in tags_dict.items():
            tags.append({"Key": name, "Value": value})
        return tags

    def _share(self, share_conf: List[str], images: Dict[str, _ImageInfo]):
        """
        Share images with accounts

        :param share_conf: the share configuration. eg. self.conf["share_create"]
        :type share_conf: List[str]
        :param images: a Dict with region names as keys and _ImageInfo objects as values
        :type images: Dict[str, _ImageInfo]
        """
        share_list: List[Dict[str, str]] = [{"UserId": user_id} for user_id in share_conf]

        for region, image_info in images.items():
            ec2client: EC2Client = boto3.client("ec2", region_name=region)
            # modify image permissions
            ec2client.modify_image_attribute(
                Attribute="LaunchPermission",
                ImageId=image_info.image_id,
                LaunchPermission={"Add": share_list},  # type: ignore
            )

            # modify snapshot permissions
            if image_info.snapshot_id:
                ec2client.modify_snapshot_attribute(
                    Attribute="createVolumePermission",
                    SnapshotId=image_info.snapshot_id,
                    CreateVolumePermission={"Add": share_list},  # type: ignore
                )

        logger.info(f"shared images & snapshots with '{share_conf}'")

    def _get_root_device_snapshot_id(self, image):
        """
        Get the root device snapshot id for a given image
        :param image: a image structure returned by eg. describe_images()["Images"][0]
        :type image: dict
        :return: Either None or a snapshot-id
        :rtype: Optional[str]
        """
        root_device_name = image.get("RootDeviceName")
        if not root_device_name:
            logger.debug(f"can not get RootDeviceName for image {image}")
            return None
        for bdm in image["BlockDeviceMappings"]:
            if bdm["DeviceName"] == root_device_name:
                ebs = bdm.get("Ebs")
                if not ebs:
                    logger.debug(
                        f"can not get RootDeviceName. root device {root_device_name} doesn't have a Ebs section"
                    )
                    return None
                logger.debug(f"found Ebs for root device {root_device_name}: {bdm['Ebs']}")
                return bdm["Ebs"]["SnapshotId"]

    def _get(self, ec2client: EC2Client) -> Optional[_ImageInfo]:
        """
        Get the a _ImageInfo for the current image which contains the ami id and
        root device snapshot id.
        This relies on the image name to be unique and will raise a MultipleImagesException
        if multiple images are found.

        :param ec2client: EC2Client
        :type ec2client: EC2Client
        :return: Either None or a _ImageInfo
        :rtype: Optional[_ImageInfo]
        """
        resp = ec2client.describe_images(
            Filters=[
                {"Name": "name", "Values": [self.image_name]},
            ],
            Owners=["self"],
        )

        if len(resp.get("Images", [])) == 1:
            root_device_snapshot_id = self._get_root_device_snapshot_id(resp["Images"][0])
            return _ImageInfo(resp["Images"][0]["ImageId"], root_device_snapshot_id)
        elif len(resp.get("Images", [])) == 0:
            return None
        else:
            images = [i["ImageId"] for i in resp.get("Images", [])]
            raise exceptions.MultipleImagesException(
                f"Found {len(images)} images ({', '.join(images)}) with "
                f"name {self.image_name} in region {ec2client.meta.region_name}. There should be only 1."
            )

    def _put_ssm_parameters(self) -> None:
        """
        Push the configured SSM parameters to the parameter store
        """
        logger.info(f"Pushing SSM parameters for image {self.image_name} in {len(self.image_regions)} regions ...")
        for region in self.image_regions:
            ec2client_region: EC2Client = boto3.client("ec2", region_name=region)
            image_info: Optional[_ImageInfo] = self._get(ec2client_region)

            # image in region not found
            if not image_info:
                logger.error(f"image {self.image_name} not available in region {region}. can not push SSM parameter")
                continue

            ssmclient_region: SSMClient = boto3.client("ssm", region_name=region)
            # iterate over all defined parameters
            for parameter in self.conf["ssm_parameter"]:
                # if overwrite is not allowed, check if the parameter is already there and if so, do nothing
                if not parameter["allow_overwrite"]:
                    resp = ssmclient_region.get_parameters(Names=[parameter["name"]])
                    if len(resp["Parameters"]) >= 1:
                        # sanity check if the available parameter matches the value we would (but don't) push
                        if resp["Parameters"][0]["Value"] != image_info.image_id:
                            logger.warning(
                                f"SSM parameter {parameter['name']} exists but value does not match "
                                f"(found {resp['Parameters'][0]['Value']}; expected: {image_info.image_id}"
                            )
                        # parameter exists already and overwrite is not allowed so continue
                        continue
                # push parameter to store
                ssmclient_region.put_parameter(
                    Name=parameter["name"],
                    Description=parameter.get("description", ""),
                    Value=image_info.image_id,
                    Type="String",
                    Overwrite=parameter["allow_overwrite"],
                    DataType="aws:ec2:image",
                    # TODO: tags can't be used together with overwrite
                    # Tags=self._ctx.tags,
                )

                logger.info(
                    f"pushed SSM parameter {parameter['name']} with value {image_info.image_id} in region {region}"
                )

    def _public(self) -> None:
        """
        Make image and underlying root device snapshot public
        """
        logger.info(f"Make image {self.image_name} in {len(self.image_regions)} regions public ...")

        for region in self.image_regions:
            ec2client_region: EC2Client = boto3.client("ec2", region_name=region)
            image_info: Optional[_ImageInfo] = self._get(ec2client_region)
            if image_info:
                ec2client_region.modify_image_attribute(
                    ImageId=image_info.image_id,
                    LaunchPermission={
                        "Add": [
                            {
                                "Group": "all",
                            },
                        ],
                    },
                )
                logger.info(f"image {image_info.image_id} in region {region} public now")

                if image_info.snapshot_id:
                    ec2client_region.modify_snapshot_attribute(
                        SnapshotId=image_info.snapshot_id,
                        Attribute="createVolumePermission",
                        GroupNames=[
                            "all",
                        ],
                        OperationType="add",
                    )
                    logger.info(
                        f"snapshot {image_info.snapshot_id} ({image_info.image_id}) in region {region} public now"
                    )
                else:
                    logger.error(
                        f"snapshot for image {self.image_name} ({image_info.image_id}) not available "
                        f"in region {region}. can not make public"
                    )
            else:
                logger.error(f"image {self.image_name} not available in region {region}. can not make public")

    def cleanup(self) -> None:
        """
        Cleanup/delete the temporary images

        If an image is marked as "temporary" in the configuration, do
        delete that image in all regions.
        Note: if a temporary image is public, it won't be deleted. A temporary
        image should never be public
        Note: the underlying snapshot is currently not deleted. That might change in
        the future
        """
        if not self.conf["temporary"]:
            logger.info(f"image {self.image_name} not marked as temporary. no cleanup")
            return

        # do the cleanup - the image is marked as temporary
        logger.info(f"Cleanup image {self.image_name} ...")
        for region in self.image_regions:
            ec2client_region: EC2Client = boto3.client("ec2", region_name=region)
            image_info: Optional[_ImageInfo] = self._get(ec2client_region)

            if image_info:
                resp = ec2client_region.describe_images(
                    Filters=[
                        {"Name": "image-id", "Values": [image_info.image_id]},
                    ]
                )
                if resp["Images"][0]["Public"] is True:
                    # this shouldn't happen because the image is marked as temporary in the config
                    # so how can it be public?
                    logger.error(
                        f"no cleanup for {self.image_name} in {region} because ({image_info.image_id}) image is public"
                    )
                else:
                    ec2client_region.deregister_image(ImageId=image_info.image_id)
                    logger.info(f"{self.image_name} in {region} ({image_info.image_id}) deleted")

    def list(self) -> Dict[str, _ImageInfo]:
        """
        Get image based on the available configuration
        This doesn't change anything - it just tries to get the available image
        for the different configured regions
        :return: a Dict with region names as keys and _ImageInfo objects as values
        :rtype: Dict[str, _ImageInfo]
        """
        images: Dict[str, _ImageInfo] = dict()
        for region in self.image_regions:
            ec2client_region: EC2Client = boto3.client("ec2", region_name=region)
            image_info: Optional[_ImageInfo] = self._get(ec2client_region)
            if image_info:
                images[region] = image_info
            else:
                logger.warning(f"image {self.image_name} not available in region {region}")
        return images

    def create(self) -> Dict[str, _ImageInfo]:
        """
        Get or create a image based on the available configuration

        :return: a Dict with region names as keys and _ImageInfo objects as values
        :rtype: Dict[str, _ImageInfo]
        """
        # this **must** be the region that is used for S3
        ec2client: EC2Client = boto3.client("ec2", region_name=self._s3.bucket_region)

        # make sure the initial snapshot exists
        self._snapshot.create(ec2client, self.snapshot_name)

        # make sure the snapshot exist in all required regions
        snapshot_ids: Dict[str, str] = self._snapshot.copy(
            self.snapshot_name, self._s3.bucket_region, self.image_regions
        )

        images: Dict[str, _ImageInfo] = dict()
        for region in self.image_regions:
            ec2client_region: EC2Client = boto3.client("ec2", region_name=region)
            image_info: Optional[_ImageInfo] = self._get(ec2client_region)
            if image_info:
                if image_info.snapshot_id != snapshot_ids[region]:
                    logger.warning(
                        f"image with name '{self.image_name}' already exists ({image_info.image_id}) "
                        f"in region {ec2client_region.meta.region_name} but the root device "
                        f"snapshot id is unexpected (got {image_info.snapshot_id} but expected {snapshot_ids[region]})"
                    )
                else:
                    logger.info(
                        f"image with name '{self.image_name}' already exists ({image_info.image_id}) "
                        f"in region {ec2client_region.meta.region_name}"
                    )
                images[region] = image_info
            else:
                logger.info(
                    f"creating image with name '{self.image_name}' in "
                    f"region {ec2client_region.meta.region_name} ..."
                )

                register_image_kwargs = dict(
                    Name=self.image_name,
                    Description=self.conf.get("description", ""),
                    Architecture=self._ctx.conf["source"]["architecture"],
                    RootDeviceName=self.conf["root_device_name"],
                    BlockDeviceMappings=[
                        {
                            "Ebs": {
                                "SnapshotId": snapshot_ids[region],
                                "VolumeType": self.conf["root_device_volume_type"],
                                "VolumeSize": self.conf["root_device_volume_size"],
                            },
                            "DeviceName": self.conf["root_device_name"],
                        },
                        # TODO: make those ephemeral block device mappings configurable
                        {"VirtualName": "ephemeral0", "DeviceName": "/dev/sdb"},
                        {"VirtualName": "ephemeral1", "DeviceName": "/dev/sdc"},
                    ],
                    EnaSupport=True,
                    SriovNetSupport="simple",
                    VirtualizationType="hvm",
                    BootMode=self.conf["boot_mode"],
                )

                if self.conf["tpm_support"]:
                    register_image_kwargs["TpmSupport"] = self.conf["tpm_support"]

                if self.conf["imds_support"]:
                    register_image_kwargs["ImdsSupport"] = self.conf["imds_support"]

                if self.conf["uefi_data"]:
                    with open(self.conf["uefi_data"], "r") as f:
                        uefi_data = f.read()
                    register_image_kwargs["UefiData"] = uefi_data

                if self.conf["billing_products"]:
                    register_image_kwargs["BillingProducts"] = self.conf["billing_products"]

                resp = ec2client_region.register_image(**register_image_kwargs)
                ec2client_region.create_tags(Resources=[resp["ImageId"]], Tags=self._tags)
                images[region] = _ImageInfo(resp["ImageId"], snapshot_ids[region])

        # wait for the images
        logger.info(f"Waiting for {len(images)} images to be ready the regions ...")
        for region, image_info in images.items():
            ec2client_region_wait: EC2Client = boto3.client("ec2", region_name=region)
            logger.info(
                f"Waiting for {image_info.image_id} in {ec2client_region_wait.meta.region_name} "
                "to exist/be available ..."
            )
            waiter_exists = ec2client_region_wait.get_waiter("image_exists")
            waiter_exists.wait(ImageIds=[image_info.image_id])
            waiter_available = ec2client_region_wait.get_waiter("image_available")
            waiter_available.wait(ImageIds=[image_info.image_id])
        logger.info(f"{len(images)} images are ready")

        # share
        if self.conf["share"]:
            self._share(self.conf["share"], images)

        return images

    def publish(self) -> None:
        """
        Handle all publication steps
        - make image and underlying root device snapshot public if the public flag is set
        - request a new marketplace version for the image in us-east-1 if the marketplace config is present
        Note: if the temporary flag is set in the image, this method will do nothing
        Note: this command doesn't unpublish anything!
        """
        # never publish temporary images
        if self.conf["temporary"]:
            logger.warning(f"image {self.image_name} marked as temporary. do not publish")
            return

        # make snapshot and image public if requested in the image
        if self.conf["public"]:
            self._public()
        else:
            logger.info(f"image {self.image_name} not marked as public. do not publish")

        # handle marketplace publication
        if self.conf["marketplace"]:
            # the "marketplace" configuration is only valid in the "aws" partition
            partition = boto3.client("ec2").meta.partition
            if partition == "aws":
                logger.info(f"marketplace version request for {self.image_name}")
                # image needs to be in us-east-1
                ec2client: EC2Client = boto3.client("ec2", region_name="us-east-1")
                image_info: Optional[_ImageInfo] = self._get(ec2client)
                if image_info:
                    im = ImageMarketplace(self._ctx, self.image_name)
                    im.request_new_version(image_info.image_id)
                else:
                    logger.error(
                        f"can not request marketplace version for {self.image_name} because no image found in us-east-1"
                    )
            else:
                logger.info(
                    f"found marketplace config for {self.image_name} and partition 'aws' but "
                    f"currently using partition {partition}. Ignoring marketplace config."
                )

        # handle SSM parameter store
        if self.conf["ssm_parameter"]:
            self._put_ssm_parameters()

    def _verify(self, region: str) -> List[ImageVerificationErrors]:
        """
        Verify (but don't modify or create anything) the image in a single region
        """
        problems: List[ImageVerificationErrors] = []
        ec2client_region: EC2Client = boto3.client("ec2", region_name=region)
        image_info: Optional[_ImageInfo] = self._get(ec2client_region)

        if not image_info:
            problems.append(ImageVerificationErrors.NOT_EXIST)
            return problems

        image_aws = ec2client_region.describe_images(ImageIds=[image_info.image_id])["Images"][0]

        # verify state
        if image_aws["State"] != "available":
            problems.append(ImageVerificationErrors.STATE_NOT_AVAILABLE)

        # verify RootDeviceType
        if image_aws["RootDeviceType"] != "ebs":
            problems.append(ImageVerificationErrors.ROOT_DEVICE_TYPE)

        # verify BootMode
        if image_aws["BootMode"] != self.conf["boot_mode"]:
            problems.append(ImageVerificationErrors.BOOT_MODE)

        # verify RootDeviceVolumeType, RootDeviceVolumeSize and Snapshot
        for bdm in image_aws["BlockDeviceMappings"]:
            if bdm.get("DeviceName") and bdm["DeviceName"] == image_aws["RootDeviceName"]:
                # here's the root device
                if bdm["Ebs"]["VolumeType"] != self.conf["root_device_volume_type"]:
                    problems.append(ImageVerificationErrors.ROOT_DEVICE_VOLUME_TYPE)
                if bdm["Ebs"]["VolumeSize"] != self.conf["root_device_volume_size"]:
                    problems.append(ImageVerificationErrors.ROOT_DEVICE_VOLUME_SIZE)

                # verify snapshot
                snapshot_aws = ec2client_region.describe_snapshots(SnapshotIds=[bdm["Ebs"]["SnapshotId"]])["Snapshots"][
                    0
                ]
                if snapshot_aws["State"] != "completed":
                    problems.append(ImageVerificationErrors.ROOT_DEVICE_SNAPSHOT_NOT_COMPLETE)

        # verify tpm support
        if self.conf["tpm_support"] and image_aws.get("TpmSupport") != self.conf["tpm_support"]:
            problems.append(ImageVerificationErrors.TPM_SUPPORT)

        # verify imds support
        if self.conf["imds_support"] and image_aws.get("ImdsSupport") != self.conf["imds_support"]:
            problems.append(ImageVerificationErrors.IMDS_SUPPORT)

        # billing products
        if self.conf["billing_products"] and image_aws.get("BillingProducts") != self.conf["billing_products"]:
            problems.append(ImageVerificationErrors.BILLING_PRODUCTS)

        # verify tags
        for tag in image_aws["Tags"]:
            if tag["Key"] == "Name" and tag["Value"] != self.snapshot_name:
                problems.append(ImageVerificationErrors.TAGS)

        return problems

    def verify(self) -> Dict[str, List[ImageVerificationErrors]]:
        """
        Verify (but don't modify or create anything) that the image configuration
        matches what is on AWS
        """
        logger.info(f"Verifying image {self.image_name} ...")
        problems: Dict[str, List[ImageVerificationErrors]] = dict()
        for region in self.image_regions:
            problems[region] = self._verify(region)

        return problems



import logging
import pathlib
from typing import Dict, Iterator, List, Optional, Tuple

from awspub.context import Context
from awspub.image import Image, _ImageInfo
from awspub.s3 import S3

logger = logging.getLogger(__name__)


def _images_grouped(
    images: List[Tuple[str, Image, Dict[str, _ImageInfo]]], group: Optional[str]
) -> Tuple[Dict[str, Dict[str, str]], Dict[str, Dict[str, Dict[str, str]]]]:
    """
    Group the given images by name and by group

    :param images: the images
    :type images: List[Tuple[str, Image, Dict[str, _ImageInfo]]]
    :param group: a optional group name
    :type group: Optional[str]
    :return: the images grouped by name and by group
    :rtype: Tuple[Dict[str, Dict[str, str]], Dict[str, Dict[str, Dict[str, str]]]
    """
    images_by_name: Dict[str, Dict[str, str]] = dict()
    images_by_group: Dict[str, Dict[str, Dict[str, str]]] = dict()
    for image_name, image, image_result in images:
        images_region_id: Dict[str, str] = {key: val.image_id for (key, val) in image_result.items()}
        images_by_name[image_name] = images_region_id
        for image_group in image.conf.get("groups", []):
            if group and image_group != group:
                continue
            if not images_by_group.get(image_group):
                images_by_group[image_group] = {}
            images_by_group[image_group][image_name] = images_region_id
    return images_by_name, images_by_group


def _images_filtered(context: Context, group: Optional[str]) -> Iterator[Tuple[str, Image]]:
    """
    Filter the images from ctx based on the given args

    :param context: the context
    :type context: a awspub.context.Context instance
    :param group: a optional group name
    :type group: Optional[str]
    """
    for image_name in context.conf["images"].keys():
        image = Image(context, image_name)
        if group:
            # limit the images to process to the group given on the command line
            if group not in image.conf.get("groups", []):
                logger.info(f"skipping image {image_name} because not part of group {group}")
                continue

        logger.info(f"processing image {image_name} (group: {group})")
        yield image_name, image


def create(
    config: pathlib.Path, config_mapping: pathlib.Path, group: Optional[str]
) -> Tuple[Dict[str, Dict[str, str]], Dict[str, Dict[str, Dict[str, str]]]]:
    """
    Create images in the partition of the used account based on
    the given configuration file and the config mapping

    :param config: the configuration file path
    :type config: pathlib.Path
    :param config_mapping: the config template mapping file path
    :type config_mapping: pathlib.Path
    :param group: only handles images from given group
    :type group: Optional[str]
    :return: the images grouped by name and by group
    :rtype: Tuple[Dict[str, Dict[str, str]], Dict[str, Dict[str, Dict[str, str]]]
    """

    ctx = Context(config, config_mapping)
    s3 = S3(ctx)
    s3.upload_file(ctx.conf["source"]["path"])
    images: List[Tuple[str, Image, Dict[str, _ImageInfo]]] = []
    for image_name, image in _images_filtered(ctx, group):
        image_result: Dict[str, _ImageInfo] = image.create()
        images.append((image_name, image, image_result))
    images_by_name, images_by_group = _images_grouped(images, group)
    return images_by_name, images_by_group


def list(
    config: pathlib.Path, config_mapping: pathlib.Path, group: Optional[str]
) -> Tuple[Dict[str, Dict[str, str]], Dict[str, Dict[str, Dict[str, str]]]]:
    """
    List images in the partition of the used account based on
    the given configuration file and the config mapping

    :param config: the configuration file path
    :type config: pathlib.Path
    :param config_mapping: the config template mapping file path
    :type config_mapping: pathlib.Path
    :param group: only handles images from given group
    :type group: Optional[str]
    :return: the images grouped by name and by group
    :rtype: Tuple[Dict[str, Dict[str, str]], Dict[str, Dict[str, Dict[str, str]]]
    """
    ctx = Context(config, config_mapping)
    images: List[Tuple[str, Image, Dict[str, _ImageInfo]]] = []
    for image_name, image in _images_filtered(ctx, group):
        image_result: Dict[str, _ImageInfo] = image.list()
        images.append((image_name, image, image_result))

    images_by_name, images_by_group = _images_grouped(images, group)
    return images_by_name, images_by_group


def publish(config: pathlib.Path, config_mapping: pathlib.Path, group: Optional[str]):
    """
    Make available images in the partition of the used account based on
    the given configuration file public

    :param config: the configuration file path
    :type config: pathlib.Path
    :param config_mapping: the config template mapping file path
    :type config_mapping: pathlib.Path
    :param group: only handles images from given group
    :type group: Optional[str]
    """
    ctx = Context(config, config_mapping)
    for image_name, image in _images_filtered(ctx, group):
        image.publish()


def cleanup(config: pathlib.Path, config_mapping: pathlib.Path, group: Optional[str]):
    """
    Cleanup available images in the partition of the used account based on
    the given configuration file

    :param config: the configuration file path
    :type config: pathlib.Path
    :param config_mapping: the config template mapping file path
    :type config_mapping: pathlib.Path
    :param group: only handles images from given group
    :type group: Optional[str]
    """
    ctx = Context(config, config_mapping)
    for image_name, image in _images_filtered(ctx, group):
        image.cleanup()


def verify(config: pathlib.Path, config_mapping: pathlib.Path, group: Optional[str]) -> Dict[str, Dict]:
    """
    Verify available images in the partition of the used account based on
    the given configuration file.
    This is EXPERIMENTAL and doesn't work reliable yet!

    :param config: the configuration file path
    :type config: pathlib.Path
    :param config_mapping: the config template mapping file path
    :type config_mapping: pathlib.Path
    :param group: only handles images from given group
    :type group: Optional[str]
    """
    problems: Dict[str, Dict] = dict()
    ctx = Context(config, config_mapping)
    for image_name, image in _images_filtered(ctx, group):
        problems[image_name] = image.verify()
    return problems



import pathlib
from unittest.mock import patch

import pytest

from awspub import context, s3
from awspub.exceptions import BucketDoesNotExistException

curdir = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize(
    "list_multipart_uploads_resp,create_multipart_upload_called",
    [
        # no available uploads - create one
        ([], True),
        # one available upload with non-matching key
        ([{"UploadId": "abc", "Key": "does-not-match"}], True),
        # multiple available upload with non-matching key
        ([{"UploadId": "abc", "Key": "does-not-match"}, {"UploadId": "def", "Key": "does-not-match2"}], True),
        # one available upload with matching key
        ([{"UploadId": "abc", "Key": "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8"}], False),
        # multiple available upload with one matching key
        (
            [
                {"UploadId": "abc", "Key": "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8"},
                {"UploadId": "abc", "Key": "does-not-match"},
            ],
            False,
        ),
        # multiple available upload with multiple matching keys
        (
            [
                {"UploadId": "abc", "Key": "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8"},
                {"UploadId": "def", "Key": "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8"},
            ],
            False,
        ),
    ],
)
def test_s3__get_multipart_upload_id(list_multipart_uploads_resp, create_multipart_upload_called):
    """
    test the _get_multipart_upload_id() function
    """

    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.list_multipart_uploads.return_value = {"Uploads": list_multipart_uploads_resp}
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        sthree = s3.S3(ctx)
        sthree._get_multipart_upload_id()
        assert instance.create_multipart_upload.called == create_multipart_upload_called


@patch("awspub.s3.S3._bucket_exists", return_value=True)
@patch("awspub.s3.boto3")
def test_s3_bucket_region_bucket_exists(boto3_mock, bucket_exists_mock):
    region_name = "sample-region-1"
    head_bucket = {"BucketRegion": region_name}
    boto3_mock.client.return_value.head_bucket.return_value = head_bucket
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    sthree = s3.S3(ctx)

    assert sthree.bucket_region == region_name


@patch("awspub.s3.S3._bucket_exists", return_value=False)
@patch("boto3.client")
def test_s3_bucket_region_bucket_not_exists(bclient_mock, bucket_exists_mock):
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    sthree = s3.S3(ctx)

    with pytest.raises(BucketDoesNotExistException):
        sthree.bucket_region()



import pathlib
from unittest.mock import patch

import pytest

from awspub import context, exceptions, image

curdir = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize(
    "imagename,snapshotname",
    [
        # test-image-1 without any separate snapshot or billing products.
        # so snapshotname should match source sha256sum
        ("test-image-1", "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8"),
        # test-image-2 with separate snapshot but without billing products.
        # so snapshotname should be the shasum of the concatenated string of:
        # - 6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8
        # - echo -n test-image-2 | sha256sum
        ("test-image-2", "0c274a96fe840cdd9cf65b0bf8e4d755d94fddf00916aa6f26ee3f08e412c88f"),
        # test-image-3 with separate snapshot and billing products
        ("test-image-3", "ef7c5bbbc2816c60acfa4f3954e431c849054f7370bf351055f6d665b60623e7"),
        # test-image-4 without separate snapshot but with billing products
        ("test-image-4", "bf795c602d53ff9c9548cc6305aa1240bd0f3d4429869abe4c96bcef65c4e48d"),
        # test-image-5 without separate snapshot but with multiple billing products
        ("test-image-5", "8171cd4d36d06150a5ff8bb519439c5efd4e91841be62f50736db3b82e4aaedc"),
    ],
)
def test_snapshot_names(imagename, snapshotname):
    """
    Test the snapshot name calculation based on the image properties
    """
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    assert ctx.conf["source"]["path"] == curdir / "fixtures/config1.vmdk"
    assert ctx.source_sha256 == "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8"

    img = image.Image(ctx, imagename)
    assert img.snapshot_name == snapshotname


@pytest.mark.parametrize(
    "imagename,regions",
    [
        # test-image-1 has 2 regions defined
        ("test-image-1", ["region1", "region2"]),
        # test-image-2 has no regions defined, so whatever the ec2 client returns should be valid
        ("test-image-2", ["all-region-1", "all-region-2"]),
    ],
)
@patch("awspub.s3.S3.bucket_region", return_value="region1")
def test_image_regions(s3_region_mock, imagename, regions):
    """
    Test the regions for a given image
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_regions.return_value = {
            "Regions": [{"RegionName": "all-region-1"}, {"RegionName": "all-region-2"}]
        }
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        img = image.Image(ctx, imagename)
        assert img.image_regions == regions


@pytest.mark.parametrize(
    "imagename,cleanup",
    [
        ("test-image-1", True),
        ("test-image-2", False),
    ],
)
def test_image_cleanup(imagename, cleanup):
    """
    Test the cleanup for a given image
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_images.return_value = {"Images": [{"Name": imagename, "Public": False, "ImageId": "ami-123"}]}
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        img = image.Image(ctx, imagename)
        img.cleanup()
        assert instance.deregister_image.called == cleanup


@pytest.mark.parametrize(
    "root_device_name,block_device_mappings,snapshot_id",
    [
        ("", [], None),
        ("/dev/sda1", [], None),
        (
            "/dev/sda1",
            [
                {
                    "DeviceName": "/dev/sda1",
                    "Ebs": {
                        "DeleteOnTermination": True,
                        "SnapshotId": "snap-0be0763f84af34e05",
                        "VolumeSize": 17,
                        "VolumeType": "gp2",
                        "Encrypted": False,
                    },
                },
                {"DeviceName": "/dev/sdb", "VirtualName": "ephemeral0"},
            ],
            "snap-0be0763f84af34e05",
        ),
    ],
)
def test_image___get_root_device_snapshot_id(root_device_name, block_device_mappings, snapshot_id):
    """
    Test the _get_root_device_snapshot_id() method
    """
    i = {"RootDeviceName": root_device_name, "BlockDeviceMappings": block_device_mappings}
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    img = image.Image(ctx, "test-image-1")
    assert img._get_root_device_snapshot_id(i) == snapshot_id


@pytest.mark.parametrize(
    "imagename,partition,called_mod_image,called_mod_snapshot,called_start_change_set,called_put_parameter",
    [
        ("test-image-6", "aws", True, True, False, False),
        ("test-image-7", "aws", False, False, False, False),
        ("test-image-8", "aws", True, True, True, True),
        ("test-image-8", "aws-cn", True, True, False, True),
    ],
)
def test_image_publish(
    imagename, partition, called_mod_image, called_mod_snapshot, called_start_change_set, called_put_parameter
):
    """
    Test the publish() for a given image
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.meta.partition = partition
        instance.describe_images.return_value = {
            "Images": [
                {
                    "Name": imagename,
                    "ImageId": "ami-abc",
                    "RootDeviceName": "/dev/sda1",
                    "BlockDeviceMappings": [
                        {
                            "DeviceName": "/dev/sda1",
                            "Ebs": {
                                "DeleteOnTermination": True,
                                "SnapshotId": "snap-0be0763f84af34e05",
                            },
                        },
                    ],
                }
            ]
        }
        instance.get_parameters.return_value = {"Parameters": []}
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        img = image.Image(ctx, imagename)
        img.publish()
        assert instance.modify_image_attribute.called == called_mod_image
        assert instance.modify_snapshot_attribute.called == called_mod_snapshot
        assert instance.start_change_set.called == called_start_change_set
        assert instance.put_parameter.called == called_put_parameter


def test_image__get_zero_images():
    """
    Test the Image._get() method with zero matching image
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_images.return_value = {"Images": []}
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        img = image.Image(ctx, "test-image-1")
        assert img._get(instance) is None


def test_image__get_one_images():
    """
    Test the Image._get() method with a single matching image
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_images.return_value = {
            "Images": [
                {
                    "Name": "test-image-1",
                    "ImageId": "ami-abc",
                    "RootDeviceName": "/dev/sda1",
                    "BlockDeviceMappings": [
                        {
                            "DeviceName": "/dev/sda1",
                            "Ebs": {
                                "DeleteOnTermination": True,
                                "SnapshotId": "snap-abc",
                            },
                        },
                    ],
                }
            ]
        }
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        img = image.Image(ctx, "test-image-1")
        assert img._get(instance) == image._ImageInfo("ami-abc", "snap-abc")


def test_image__get_multiple_images():
    """
    Test the Image._get() method with a multiple matching image
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_images.return_value = {
            "Images": [
                {
                    "Name": "test-image-1",
                    "ImageId": "ami-1,",
                },
                {
                    "Name": "test-image-1",
                    "ImageId": "ami-2,",
                },
            ]
        }
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        img = image.Image(ctx, "test-image-1")
        with pytest.raises(exceptions.MultipleImagesException):
            img._get(instance)


@pytest.mark.parametrize(
    "imagename,expected_tags",
    [
        # no image specific tags - assume the common tags
        (
            "test-image-1",
            [
                {"Key": "awspub:source:filename", "Value": "config1.vmdk"},
                {"Key": "awspub:source:architecture", "Value": "x86_64"},
                {
                    "Key": "awspub:source:sha256",
                    "Value": "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8",
                },
                {"Key": "name", "Value": "foobar"},
            ],
        ),
        # with image specific tag but no override
        (
            "test-image-6",
            [
                {"Key": "awspub:source:filename", "Value": "config1.vmdk"},
                {"Key": "awspub:source:architecture", "Value": "x86_64"},
                {
                    "Key": "awspub:source:sha256",
                    "Value": "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8",
                },
                {"Key": "name", "Value": "foobar"},
                {"Key": "key1", "Value": "value1"},
            ],
        ),
        # with image specific tag which overrides common tag
        (
            "test-image-7",
            [
                {"Key": "awspub:source:filename", "Value": "config1.vmdk"},
                {"Key": "awspub:source:architecture", "Value": "x86_64"},
                {
                    "Key": "awspub:source:sha256",
                    "Value": "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8",
                },
                {"Key": "name", "Value": "not-foobar"},
                {"Key": "key2", "Value": "name"},
            ],
        ),
    ],
)
def test_image__tags(imagename, expected_tags):
    """
    Test the Image._tags() method
    """
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    img = image.Image(ctx, imagename)
    assert img._tags == expected_tags


@pytest.mark.parametrize(
    "available_images,expected",
    [
        # image available
        ([{"Name": "test-image-6", "ImageId": "ami-123"}], {"eu-central-1": image._ImageInfo("ami-123", None)}),
        # image not available
        ([], {}),
    ],
)
def test_image_list(available_images, expected):
    """
    Test the list for a given image
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_images.return_value = {"Images": available_images}
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        img = image.Image(ctx, "test-image-6")
        assert img.list() == expected


@patch("awspub.s3.S3.bucket_region", return_value="region1")
def test_image_create_existing(s3_bucket_mock):
    """
    Test the create() method for a given image that already exist
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_snapshots.return_value = {"Snapshots": [{"SnapshotId": "snap-123"}]}
        instance.describe_images.return_value = {
            "Images": [
                {
                    "Name": "test-image-6",
                    "ImageId": "ami-123",
                    "RootDeviceName": "/dev/sda1",
                    "BlockDeviceMappings": [
                        {
                            "DeviceName": "/dev/sda1",
                            "Ebs": {
                                "DeleteOnTermination": True,
                                "SnapshotId": "snap-123",
                            },
                        },
                    ],
                }
            ]
        }
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        img = image.Image(ctx, "test-image-6")
        assert img.create() == {"eu-central-1": image._ImageInfo(image_id="ami-123", snapshot_id="snap-123")}
        # register and create_tags shouldn't be called given that the image was already there
        assert not instance.register_image.called
        assert not instance.create_tags.called


@pytest.mark.parametrize(
    "imagename,describe_images,get_parameters,get_parameters_called,put_parameter_called",
    [
        # no image, no parameters (this should actually never happen but test it anyway)
        ("test-image-8", [], [], False, False),
        # with image, no parameter, no overwrite
        ("test-image-8", [{"Name": "test-image-8", "ImageId": "ami-123"}], [], True, True),
        # with image, with parameter, no overwrite
        (
            "test-image-8",
            [{"Name": "test-image-8", "ImageId": "ami-123"}],
            [{"Name": "/awspub-test/param1", "Value": "ami-123"}],
            True,
            False,
        ),
        # with image, no parameter, with overwrite
        ("test-image-9", [{"Name": "test-image-8", "ImageId": "ami-123"}], [], False, True),
        # with image, with parameter, with overwrite
        (
            "test-image-9",
            [{"Name": "test-image-8", "ImageId": "ami-123"}],
            [{"Name": "/awspub-test/param1", "Value": "ami-123"}],
            False,
            True,
        ),
    ],
)
def test_image__put_ssm_parameters(
    imagename, describe_images, get_parameters, get_parameters_called, put_parameter_called
):
    """
    Test the _put_ssm_parameters() method
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_images.return_value = {"Images": describe_images}
        instance.get_parameters.return_value = {"Parameters": get_parameters}
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        img = image.Image(ctx, imagename)
        img._put_ssm_parameters()
        assert instance.get_parameters.called == get_parameters_called
        assert instance.put_parameter.called == put_parameter_called


@pytest.mark.parametrize(
    "image_found,config,config_image_name, expected_problems",
    [
        # image not available
        ([], "fixtures/config1.yaml", "test-image-6", [image.ImageVerificationErrors.NOT_EXIST]),
        # image matches expectations from config
        (
            [
                {
                    "Name": "test-image-6",
                    "State": "available",
                    "ImageId": "ami-123",
                    "RootDeviceName": "/dev/sda1",
                    "RootDeviceType": "ebs",
                    "BootMode": "uefi-preferred",
                    "BlockDeviceMappings": [
                        {
                            "DeviceName": "/dev/sda1",
                            "Ebs": {
                                "DeleteOnTermination": True,
                                "VolumeType": "gp3",
                                "VolumeSize": 8,
                                "SnapshotId": "snap-123",
                            },
                        },
                    ],
                    "Tags": [
                        {"Key": "name", "Value": "foobar"},
                    ],
                }
            ],
            "fixtures/config1.yaml",
            "test-image-6",
            [],
        ),
    ],
)
def test_image__verify(image_found, config, config_image_name, expected_problems):
    """
    Test _verify() for a given image and configuration
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_images.return_value = {"Images": image_found}
        instance.describe_snapshots.return_value = {"Snapshots": [{"State": "completed"}]}
        ctx = context.Context(curdir / config, None)
        img = image.Image(ctx, config_image_name)
        problems = img._verify("eu-central-1")
        assert problems == expected_problems



import pathlib
from typing import Dict, List, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field


class ConfigS3Model(BaseModel):
    """
    S3 configuration.
    This is required for uploading source files (usually .vmdk) to a bucket so
    snapshots can be created out of the s3 file
    """

    model_config = ConfigDict(extra="forbid")

    bucket_name: str = Field(description="The S3 bucket name")


class ConfigSourceModel(BaseModel):
    """
    Source configuration.
    This defines the source (usually .vmdk) that is uploaded
    to S3 and then used to create EC2 snapshots in different regions.
    """

    model_config = ConfigDict(extra="forbid")

    path: pathlib.Path = Field(description="Path to a local .vmdk image")
    architecture: Literal["x86_64", "arm64"] = Field(description="The architecture of the given .vmdk image")


class ConfigImageMarketplaceSecurityGroupModel(BaseModel):
    """
    Image/AMI Marketplace specific configuration for a security group
    """

    model_config = ConfigDict(extra="forbid")

    from_port: int = Field(description="The source port")
    ip_protocol: Literal["tcp", "udp"] = Field(description="The IP protocol (either 'tcp' or 'udp')")
    ip_ranges: List[str] = Field(description="IP ranges to allow, in CIDR format (eg. '192.0.2.0/24')")
    to_port: int = Field(description="The destination port")


class ConfigImageMarketplaceModel(BaseModel):
    """
    Image/AMI Marketplace specific configuration to request new Marketplace versions
    See https://docs.aws.amazon.com/marketplace-catalog/latest/api-reference/ami-products.html
    for further information
    """

    model_config = ConfigDict(extra="forbid")

    entity_id: str = Field(description="The entity ID (product ID)")
    # see https://docs.aws.amazon.com/marketplace/latest/userguide/ami-single-ami-products.html#single-ami-marketplace-ami-access  # noqa:E501
    access_role_arn: str = Field(
        description="IAM role Amazon Resource Name (ARN) used by AWS Marketplace to access the provided AMI"
    )
    version_title: str = Field(description="The version title. Must be unique across the product")
    release_notes: str = Field(description="The release notes")
    user_name: str = Field(description="The login username to access the operating system")
    scanning_port: int = Field(description="Port to access the operating system (default: 22)", default=22)
    os_name: str = Field(description="Operating system name displayed to Marketplace buyers")
    os_version: str = Field(description="Operating system version displayed to Marketplace buyers")
    usage_instructions: str = Field(
        description=" Instructions for using the AMI, or a link to more information about the AMI"
    )
    recommended_instance_type: str = Field(
        description="The instance type that is recommended to run the service with the AMI and is the "
        "default for 1-click installs of your service"
    )
    security_groups: Optional[List[ConfigImageMarketplaceSecurityGroupModel]]


class ConfigImageSSMParameterModel(BaseModel):
    """
    Image/AMI SSM specific configuration to push parameters of type `aws:ec2:image` to the parameter store
    """

    model_config = ConfigDict(extra="forbid")

    name: str = Field(
        description="The fully qualified name of the parameter that you want to add to the system. "
        "A parameter name must be unique within an Amazon Web Services Region"
    )
    description: Optional[str] = Field(
        description="Information about the parameter that you want to add to the system", default=None
    )
    allow_overwrite: Optional[bool] = Field(
        description="allow to overwrite an existing parameter. Useful for keep a 'latest' parameter (default: false)",
        default=False,
    )


class ConfigImageModel(BaseModel):
    """
    Image/AMI configuration.
    """

    model_config = ConfigDict(extra="forbid")

    description: Optional[str] = Field(description="Optional image description", default=None)
    regions: Optional[List[str]] = Field(
        description="Optional list of regions for this image. If not given, all available regions will be used",
        default=None,
    )
    separate_snapshot: bool = Field(description="Use a separate snapshot for this image?", default=False)
    billing_products: Optional[List[str]] = Field(description="Optional list of billing codes", default=None)
    boot_mode: Literal["legacy-bios", "uefi", "uefi-preferred"] = Field(
        description="The boot mode. For arm64, this needs to be 'uefi'"
    )
    root_device_name: Optional[str] = Field(description="The root device name", default="/dev/sda1")
    root_device_volume_type: Optional[Literal["gp2", "gp3"]] = Field(
        description="The root device volume type", default="gp3"
    )
    root_device_volume_size: Optional[int] = Field(description="The root device volume size (in GB)", default=8)
    uefi_data: Optional[pathlib.Path] = Field(
        description="Optional path to a non-volatile UEFI variable store (must be already base64 encoded)", default=None
    )
    tpm_support: Optional[Literal["v2.0"]] = Field(
        description="Optional TPM support. If this is set, 'boot_mode' must be 'uefi'", default=None
    )
    imds_support: Optional[Literal["v2.0"]] = Field(description="Optional IMDS support", default=None)
    share: Optional[List[str]] = Field(
        description="Optional list of account IDs the image and snapshot will be shared with", default=None
    )
    temporary: Optional[bool] = Field(
        description="Optional boolean field indicates that a image is only temporary", default=False
    )
    public: Optional[bool] = Field(
        description="Optional boolean field indicates if the image should be public", default=False
    )
    marketplace: Optional[ConfigImageMarketplaceModel] = Field(
        description="Optional structure containing Marketplace related configuration for the commercial "
        "'aws' partition",
        default=None,
    )
    ssm_parameter: Optional[List[ConfigImageSSMParameterModel]] = Field(
        description="Optional list of SSM parameter paths of type `aws:ec2:image` which will "
        "be pushed to the parameter store",
        default=None,
    )
    groups: Optional[List[str]] = Field(description="Optional list of groups this image is part of", default=[])
    tags: Optional[Dict[str, str]] = Field(description="Optional Tags to apply to this image only", default={})


class ConfigModel(BaseModel):
    """
    The base model for the whole configuration
    """

    model_config = ConfigDict(extra="forbid")

    s3: ConfigS3Model
    source: ConfigSourceModel
    images: Dict[str, ConfigImageModel]
    tags: Optional[Dict[str, str]] = Field(description="Optional Tags to apply to all resources", default={})



import pathlib
from unittest.mock import patch

import pytest

from awspub import context, image_marketplace

curdir = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize(
    "imagename,new_version,called_start_change_set",
    [
        # same version that already exists
        ("test-image-8", "1.0.0", False),
        # new version
        ("test-image-8", "2.0.0", True),
    ],
)
def test_image_marketplace_request_new_version(imagename, new_version, called_start_change_set):
    """
    Test the request_new_version logic
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_entity.return_value = {"DetailsDocument": {"Versions": [{"VersionTitle": new_version}]}}
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        img = image_marketplace.ImageMarketplace(ctx, imagename)
        img.request_new_version("ami-123")
        assert instance.start_change_set.called == called_start_change_set


@pytest.mark.parametrize(
    "name,expected",
    [
        ("1.0.0", "1.0.0"),
        ("1.0.0 (testing)", "1.0.0 testing"),
        ("a sentence with spaces", "a sentence with spaces"),
        ("_+=.:@-", "_+=.:@-"),
        ("(parens) [brackets] |pipes|", "parens brackets pipes"),
    ],
)
def test_changeset_name_sanitization(name, expected):
    assert image_marketplace.ImageMarketplace.sanitize_changeset_name(name) == expected



import pathlib

import pytest

from awspub import api, context, image

curdir = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize(
    "group,expected_image_names",
    [
        # without any group, all images should be processed
        (
            None,
            [
                "test-image-1",
                "test-image-2",
                "test-image-3",
                "test-image-4",
                "test-image-5",
                "test-image-6",
                "test-image-7",
                "test-image-8",
                "test-image-9",
            ],
        ),
        # with a group that no image as, no image should be processed
        (
            "group-not-used",
            [],
        ),
        # with a group that an image has
        (
            "group2",
            ["test-image-1"],
        ),
        # with a group that multiple images have
        (
            "group1",
            ["test-image-1", "test-image-2"],
        ),
    ],
)
def test_api__images_filtered(group, expected_image_names):
    """
    Test the _images_filtered() function
    """
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)

    image_names = [i[0] for i in api._images_filtered(ctx, group)]
    assert image_names == expected_image_names


@pytest.mark.parametrize(
    "group,expected",
    [
        # without any group, all images should be processed
        (
            None,
            (
                {"test-image-1": {"eu-central-1": "ami-123", "eu-central-2": "ami-456"}},
                {
                    "group1": {"test-image-1": {"eu-central-1": "ami-123", "eu-central-2": "ami-456"}},
                    "group2": {"test-image-1": {"eu-central-1": "ami-123", "eu-central-2": "ami-456"}},
                },
            ),
        ),
        # with a group that no image as, image should be there but nothing in the group
        ("group-not-used", ({"test-image-1": {"eu-central-1": "ami-123", "eu-central-2": "ami-456"}}, {})),
    ],
)
def test_api__images_grouped(group, expected):
    """
    Test the _images_grouped() function
    """
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    images = [
        (
            "test-image-1",
            image.Image(ctx, "test-image-1"),
            {"eu-central-1": image._ImageInfo("ami-123", None), "eu-central-2": image._ImageInfo("ami-456", None)},
        )
    ]
    grouped = api._images_grouped(images, group)
    assert grouped == expected



import logging
import re
from typing import Any, Dict

import boto3
from mypy_boto3_marketplace_catalog import MarketplaceCatalogClient

from awspub.context import Context

logger = logging.getLogger(__name__)


class ImageMarketplace:
    """
    Handle AWS Marketplace API interaction
    """

    def __init__(self, context: Context, image_name: str):
        self._ctx: Context = context
        self._image_name: str = image_name
        # marketplace-catalog API is only available via us-east-1
        self._mpclient: MarketplaceCatalogClient = boto3.client("marketplace-catalog", region_name="us-east-1")

    @property
    def conf(self) -> Dict[str, Any]:
        """
        The marketplace configuration for the current image (based on "image_name") from context
        """
        return self._ctx.conf["images"][self._image_name]["marketplace"]

    def request_new_version(self, image_id: str) -> None:
        """
        Request a new Marketplace version for the given image Id

        :param image_id: an image Id (in the format 'ami-123')
        :type image_id: str
        """
        entity = self._mpclient.describe_entity(Catalog="AWSMarketplace", EntityId=self.conf["entity_id"])
        # check if the version already exists
        for version in entity["DetailsDocument"]["Versions"]:
            if version["VersionTitle"] == self.conf["version_title"]:
                logger.info(f"Marketplace version '{self.conf['version_title']}' already exists. Do nothing")
                return

        # version doesn't exist already - create a new one
        changeset = self._request_new_version_changeset(image_id)
        changeset_name = ImageMarketplace.sanitize_changeset_name(
            f"New version request for {self.conf['version_title']}"
        )
        resp = self._mpclient.start_change_set(
            Catalog="AWSMarketplace", ChangeSet=changeset, ChangeSetTags=self._ctx.tags, ChangeSetName=changeset_name
        )
        logger.info(
            f"new version '{self.conf['version_title']}' (image: {image_id}) for entity "
            f"{self.conf['entity_id']} requested (changeset-id: {resp['ChangeSetId']})"
        )

    def _request_new_version_changeset(self, image_id: str):
        """
        Create a changeset structure for a new AmiProduct version
        See https://docs.aws.amazon.com/marketplace-catalog/latest/api-reference/ami-products.html#ami-add-version

        :param image_id: an image Id (in the format 'ami-123')
        :type image_id: str
        :return: A changeset structure to request a new version
        :rtype: List[Dict[str, Any]]
        """
        return [
            {
                "ChangeType": "AddDeliveryOptions",
                "Entity": {
                    "Identifier": self.conf["entity_id"],
                    "Type": "AmiProduct@1.0",
                },
                "DetailsDocument": {
                    "Version": {
                        "VersionTitle": self.conf["version_title"],
                        "ReleaseNotes": self.conf["release_notes"],
                    },
                    "DeliveryOptions": [
                        {
                            "Details": {
                                "AmiDeliveryOptionDetails": {
                                    "AmiSource": {
                                        "AmiId": image_id,
                                        "AccessRoleArn": self.conf["access_role_arn"],
                                        "UserName": self.conf["user_name"],
                                        "OperatingSystemName": self.conf["os_name"],
                                        "OperatingSystemVersion": self.conf["os_version"],
                                    },
                                    "UsageInstructions": self.conf["usage_instructions"],
                                    "RecommendedInstanceType": self.conf["recommended_instance_type"],
                                    "SecurityGroups": [
                                        {
                                            "IpProtocol": sg["ip_protocol"],
                                            "IpRanges": [ipr for ipr in sg["ip_ranges"]],
                                            "FromPort": sg["from_port"],
                                            "ToPort": sg["to_port"],
                                        }
                                        for sg in self.conf["security_groups"]
                                    ],
                                }
                            }
                        }
                    ],
                },
            }
        ]

    @staticmethod
    def sanitize_changeset_name(name: str) -> str:
        # changeset names can only include alphanumeric characters, whitespace, and any combination of the following
        # characters: _+=.:@- This regex pattern takes the list of allowed characters, does a negative match on the
        # string and removes all matched (i.e. disallowed) characters. See [0] for reference.
        # [0] https://docs.aws.amazon.com/marketplace-catalog/latest/api-reference/API_StartChangeSet.html#API_StartChangeSet_RequestSyntax  # noqa
        return re.sub(
            "[^\\w\\s+=.:@-]",
            "",
            name,
        )
